# coding=utf-8
import jwt
import requests


class CodeGrantFlowTokenResponse:
    """
    Container/"DTO"-of-sorts for (persistent) token response properties, helps with IDEs.
    """
    def __init__(self, access_token, token_type, user_id, organization_id, app_id):
        self.access_token = access_token  # "Persistent" access token / "API KEY"
        self.token_type = token_type  # Keyword to use in "Authorization" Header as a prefix for the token
        self.user_id = user_id  # ID of the user to which the token grants an access
        self.organization_id = organization_id  # ID of the organization in which context the token allows API requests
        self.app_id = app_id  # ID of (your) app that requested for persistent access token


class GiosgTriggerInDjango:
    """
    Abstraction for views in Django Framework, to conveniently validate and parse Giosg Trigger Requests.
     - Validates query parameters of the trigger request
     - Validates the signature of data-JWT using app secret
     - Validates expiration times of both data-JWT and token-JWT
     - Validates existence of redirect_uri if type is "install"
     - Raises ValueError with an error message if any check fails.

     tldr: Validates everything for you and raises ValueError in case of errors. Provides helper methods.

    Additionally provides methods

        get_auth_header_content():
            constructs a string for Authorization header content, using trigger's token and app secret
            returns string to insert in "Authorization" Header, to query Giosg HTTP API while valid (.access_token_exp)
            https://developers.giosg.com/giosg_apps.html#using-http-api-from-app-triggers

        exchange_persistent_token(): ONLY VIABLE WITH BOT USER -- ELSE RAISES A VALUE ERROR
            requests (via Code Grant Flow) for a persistent access token
            returns CodeGrantFlowTokenResponse, or raises ValueError with an error message
            https://developers.giosg.com/authentication.html#token-request-with-authorization-code-grant-flow

     Properties:

        .type -- string : Trigger type
        .user_id -- string : UUID of the current user or null if not applicable
        .org_id -- string : UUID of the organization/company where the user_id belongs to
        .chat_id -- string : UUID of the related chat, if any, otherwise null
        .visitor_id -- string : ID of the related visitor, if available, otherwise null
        .inst_id -- string : UUID of the app instance/installation.
        .app_id -- string : UUID of the app
        .app_user_id -- string : UUID of any related bot user, or null if no such user was created on app installation
        .access_token -- string : Non-persistent JWT (JSON Web Token) to use when using giosg HTTP API endpoints
        .access_token_exp -- string : Non-persistent
    """
    def __init__(self, django_request, giosg_app_secret):
        """
        :param django_request: https://docs.djangoproject.com/en/2.1/ref/request-response/#httprequest-objects
        :param giosg_app_secret: https://developers.giosg.com/giosg_apps.html#trigger-requests
        """

        # Check type of the trigger condition exists
        trigger_type = django_request.GET.get('type')
        if not trigger_type:
            raise ValueError('Missing URI query parameter "type"')

        # Check data JWT (signed with the app secret) exists
        app_data = django_request.GET.get('data')
        if not app_data:
            raise ValueError('Missing URI query parameter "data"')

        # Check request is coming from Giosg (via signed data-JWT), and check both JWTs for expiration
        try:
            app_data = jwt.decode(app_data, giosg_app_secret, algorithm='HS256')
        except jwt.InvalidSignatureError:
            raise ValueError('Data-JWT has an invalid signature, your connection may be tampered.')
        except jwt.ExpiredSignatureError:
            raise ValueError('Data-JWT is expired, request may be replayed from browser history (or tampered).')

        # Additionally check redirect URI exists if type is "install", and assign it to properties if so
        self.redirect_uri = None
        if trigger_type == 'install':
            self.redirect_uri = django_request.GET.get('redirect_uri')
            if not self.redirect_uri:
                raise ValueError('Missing URI query parameter "redirect_uri"')

        # Additionally check access token if type is ANYTHING BUT uninstall, and assign it to properties if so
        self.access_token = None
        self.access_token_exp = None
        if trigger_type != 'uninstall':
            access_token = django_request.GET.get('token')
            if not access_token:
                raise ValueError('Missing URI query parameter "token"')
            access_token_data = jwt.decode(access_token, verify=False)
            self.access_token = access_token
            self.access_token_exp = access_token_data['exp']

        # Assign ever-present properties
        self.type = trigger_type
        self.user_id = app_data['user_id']
        self.org_id = app_data['org_id']
        self.chat_id = app_data['chat_id']
        self.visitor_id = app_data['visitor_id']
        self.inst_id = app_data['inst_id']
        self.app_id = app_data['app_id']
        self.app_user_id = app_data['app_user_id']

        # Set these "private" for methods
        self.__app_secret = giosg_app_secret
        self.__app_user_code = app_data['app_user_code']

    def get_auth_header_content(self):
        """
        :return: String for "Authorization" HTTP header's content
        """
        # Check the app trigger for access token, it is NOT present if type is 'uninstall'
        if not self.access_token:
            raise ValueError('An access token is required to form an Authorization header.')
        return 'GIOSGAPP {} {}'.format(self.access_token, self.__app_secret)

    def exchange_persistent_access_token(self):
        """
        :return: CodeGrantFlowTokenResponse with persistent access token
        """
        # Check the app trigger for an existing bot user, else it is impossible to request for persistent token
        if not self.app_user_id:
            raise ValueError('A bot user is required to request a persistent access token.')

        # Default Content-Type is application/x-www-form-urlencoded as required by Giosg API
        r = requests.post(
            'https://service.giosg.com/identity/token',
            data={
                'grant_type': 'authorization_code',
                'code': self.__app_user_code,
                'client_id': self.app_id,
                'client_secret': self.__app_secret
            }
        )

        # Unpack response content
        r_data = r.json()

        # Process any (known) errors
        if r.status_code == 400:
            raise ValueError('{} {}'.format(r_data['error'], r_data['error_description']))

        return CodeGrantFlowTokenResponse(
            r_data['access_token'],
            r_data['token_type'],
            r_data['user_id'],
            r_data['organization_id'],
            r_data['app_id']
        )



from django.test import TestCase, RequestFactory
from ..lib.giosg_trigger_in_django import GiosgTriggerInDjango
import jwt
import time
import os

SIGN_KEY = '123abc456cde'
class GiosgTriggerAbstractionTest(TestCase):

    @classmethod
    def setUpTestData(cls):
        # Use old JWTs from test app, unpack them without validating signature since we sign the data again anyway
        # To-Consider: Building the sample JWTs from mere dictionary data set
        data_jwt = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VyX2lkIjoiMDAwZjcxMTYtNDQ5OS0xMWU5LWJjYTQtMDI0MmFjMTEwMDA5Iiwic3ViIjoic2VydmljZS5naW9zZy5jb20iLCJvcmdfaWQiOiJhMTdjZWE4MC1lMzk3LTExZTAtYjUxYS0wMDE2M2UwYzAxZjIiLCJhcHBfdXNlcl9jb2RlIjoiZXlKaGJHY2lPaUpJVXpJMU5pSXNJblI1Y0NJNklrcFhWQ0o5LmV5SnZjbWRoYm1sNllYUnBiMjVmYVdRaU9pSmhNVGRqWldFNE1DMWxNemszTFRFeFpUQXRZalV4WVMwd01ERTJNMlV3WXpBeFpqSWlMQ0pxZEdraU9pSTRNMkl6T0dSbVlTMDFNak5tTFRFeFpUa3RPREExT0Mwd01qUXlZV014TVRBd01EZ2lMQ0poZFdRaU9sc2lhSFIwY0hNNkx5OXpaWEoyYVdObExtZHBiM05uTG1OdmJTOXBaR1Z1ZEdsMGVTOTBiMnRsYmlKZExDSmxlSEFpT2pFMU5UTTROelk0TmpVdU56VXhNVEV4TENKMWMyVnlYMmxrSWpvaU9ETTVPVFZoTlRJdE5USXpaaTB4TVdVNUxUZ3dOVGd0TURJME1tRmpNVEV3TURBNElpd2lhWE56SWpvaWFIUjBjSE02THk5elpYSjJhV05sTG1kcGIzTm5MbU52YlM5cFpHVnVkR2wwZVM5aGRYUm9iM0pwZW1VaUxDSnBZWFFpT2pFMU5UTTROelk0TXpVdU56VXhNVEVzSW1Gd2NGOXBaQ0k2SWpKa05ERTNaVEl5TFRSbVpEUXRNVEZsT1MwNU9USTRMVEF5TkRKaFl6RXhNREF3TlNKOS5PUlkta1FDLTNBNExyczFVMDBHMF9vblpPUlRia2tjY1dWRkZfUGdiTkFvIiwidmlzaXRvcl9pZCI6IiIsImFwcF9pZCI6IjJkNDE3ZTIyLTRmZDQtMTFlOS05OTI4LTAyNDJhYzExMDAwNSIsImluc3RfaWQiOiI4MzllZGE3Yy01MjNmLTExZTktODA1OC0wMjQyYWMxMTAwMDgiLCJhcHBfdXNlcl9pZCI6IjgzOTk1YTUyLTUyM2YtMTFlOS04MDU4LTAyNDJhYzExMDAwOCIsImNoYXRfaWQiOiIiLCJleHAiOjE1NTM4Nzc0MzUsInJvb21faWQiOm51bGwsIl9vcmdfaWQiOjEsImRvbWFpbl9ob3N0IjpudWxsLCJfdXNlcl9pZCI6Mjg4NDJ9.ST7ab1dvKuiKcsRI677o5CgK_t3Qrb4PLQNoXxy5T2U'
        access_token_jwt = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VyX2lkIjoiMDAwZjcxMTYtNDQ5OS0xMWU5LWJjYTQtMDI0MmFjMTEwMDA5Iiwic3ViIjoic2VydmljZS5naW9zZy5jb20iLCJwZXJtcyI6W10sIm9yZ19pZCI6ImExN2NlYTgwLWUzOTctMTFlMC1iNTFhLTAwMTYzZTBjMDFmMiIsImFwcF9pZCI6IjJkNDE3ZTIyLTRmZDQtMTFlOS05OTI4LTAyNDJhYzExMDAwNSIsImluc3RfaWQiOiI4MzllZGE3Yy01MjNmLTExZTktODA1OC0wMjQyYWMxMTAwMDgiLCJleHAiOjE1NTM4Nzc0MzV9.Rj83FyBEXS_sTQxHgq_WDjhUD3C80e-5lw1MgyczNuo'
        data_content = jwt.decode(data_jwt, verify=False)
        access_token_content = jwt.decode(access_token_jwt, verify=False)
        # Create valid JWTs
        data_content['exp'] = int(time.time()) + 60*10  # Expiration in 10 minutes since set-up of TestCase
        access_token_content['exp'] = int(time.time()) + 60*10
        cls.VALID_DATA_JWT = jwt.encode(data_content, SIGN_KEY, algorithm='HS256')
        cls.VALID_TOKEN_JWT = jwt.encode(access_token_content, SIGN_KEY, algorithm='HS256')
        # Create forged (invalid signing key) data-JWT, access token JWT is not signed
        cls.FORGED_DATA_JWT = jwt.encode(data_content, 'fake_key_lol', algorithm='HS256')
        # Create expired JWTs, access token JWT is not checked -- we explicitly use data-JWT for authorization
        data_content['exp'] = int(time.time()) - 60*10
        cls.EXPIRED_DATA_JWT = jwt.encode(data_content, SIGN_KEY, algorithm='HS256')

    def test_missing_trigger_type(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            # Missing param 'type'
            'data': self.VALID_DATA_JWT,
            'token': self.VALID_TOKEN_JWT
        })
        try:
            trigger = GiosgTriggerInDjango(request, SIGN_KEY)
            # Should go to the exception instead
            self.fail()
        except ValueError as e:
            self.assertTrue('type' in str(e))

    def test_missing_data_jwt(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            'type': 'setup',
            # Missing param 'data'
            'token': self.VALID_TOKEN_JWT
        })
        try:
            trigger = GiosgTriggerInDjango(request, SIGN_KEY)
            # Should go to the exception instead
            self.fail()
        except ValueError as e:
            self.assertTrue('data' in str(e))

    def test_invalid_signature_in_data_jwt(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            'type': 'setup',
            'data': self.FORGED_DATA_JWT,
            'token': self.VALID_TOKEN_JWT
        })
        try:
            trigger = GiosgTriggerInDjango(request, SIGN_KEY)
            # Should go to the exception instead
            self.fail()
        except ValueError as e:
            self.assertTrue('signature' in str(e))

    def test_expired_signature_in_data_jwt(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            'type': 'setup',
            'data': self.EXPIRED_DATA_JWT,
            'token': self.VALID_TOKEN_JWT
        })
        try:
            trigger = GiosgTriggerInDjango(request, SIGN_KEY)
            # Should go to the exception instead
            self.fail()
        except ValueError as e:
            self.assertTrue('expired' in str(e))

    def test_missing_install_redirect_uri(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            'type': 'install',
            'data': self.VALID_DATA_JWT,
            'token': self.VALID_TOKEN_JWT
            # Missing 'redirect_uri' when type is 'install'
        })
        try:
            trigger = GiosgTriggerInDjango(request, SIGN_KEY)
            # Should go to the exception instead
            self.fail()
        except ValueError as e:
            self.assertTrue('redirect_uri' in str(e))

    def test_missing_token_when_not_uninstall(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            'type': 'setup',
            'data': self.VALID_DATA_JWT
            # Missing param 'token'
        })
        try:
            trigger = GiosgTriggerInDjango(request, SIGN_KEY)
            # Should go to the exception instead
            self.fail()
        except ValueError as e:
            self.assertTrue('token' in str(e))

    def test_accessible_properties(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            'type': 'setup',
            'data': self.VALID_DATA_JWT,
            'token': self.VALID_TOKEN_JWT
        })
        trigger = GiosgTriggerInDjango(request, SIGN_KEY)

        # Check attrs exist
        self.assertTrue(hasattr(trigger, 'type'))
        self.assertTrue(hasattr(trigger, 'user_id'))
        self.assertTrue(hasattr(trigger, 'org_id'))
        self.assertTrue(hasattr(trigger, 'chat_id'))
        self.assertTrue(hasattr(trigger, 'visitor_id'))
        self.assertTrue(hasattr(trigger, 'inst_id'))
        self.assertTrue(hasattr(trigger, 'app_id'))
        self.assertTrue(hasattr(trigger, 'app_user_id'))
        self.assertTrue(hasattr(trigger, 'redirect_uri'))
        self.assertTrue(hasattr(trigger, 'access_token'))
        self.assertTrue(hasattr(trigger, 'access_token_exp'))

        # Check attrs are correctly typed
        self.assertTrue(isinstance(trigger.type, str))
        self.assertTrue(isinstance(trigger.user_id, str) or trigger.user_id is None)
        self.assertTrue(isinstance(trigger.org_id, str))
        self.assertTrue(isinstance(trigger.chat_id, str) or trigger.chat_id is None)
        self.assertTrue(isinstance(trigger.visitor_id, str) or trigger.visitor_id is None)
        self.assertTrue(isinstance(trigger.inst_id, str))
        self.assertTrue(isinstance(trigger.app_id, str))
        self.assertTrue(isinstance(trigger.app_user_id, str) or trigger.app_user_id is None)
        self.assertTrue(isinstance(trigger.redirect_uri, str) or trigger.redirect_uri is None)
        self.assertTrue(isinstance(trigger.access_token, str) or trigger.access_token is None)
        self.assertTrue(isinstance(trigger.access_token_exp, int) or trigger.access_token_exp is None)

    def test_authorization_header_content_getter(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            'type': 'setup',
            'data': self.VALID_DATA_JWT,
            'token': self.VALID_TOKEN_JWT
        })
        trigger = GiosgTriggerInDjango(request, SIGN_KEY)
        self.assertTrue(isinstance(trigger.get_auth_header_content(), str))

    def test_auth_header_content_getter_unavailable_if_uninstall(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            'type': 'setup',
            'data': self.VALID_DATA_JWT,
            'token': self.VALID_TOKEN_JWT
        })
        try:
            trigger = GiosgTriggerInDjango(request, SIGN_KEY)
            trigger.get_auth_header_content()
        except ValueError as e:
            self.assertTrue('token' in str(e))

    def test_persistent_token_exchanger_when_no_bot_user(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            'type': 'setup',
            'data': self.VALID_DATA_JWT,
            'token': self.VALID_TOKEN_JWT
        })
        trigger = GiosgTriggerInDjango(request, SIGN_KEY)
        trigger.app_user_id = None
        try:
            trigger.exchange_persistent_access_token()
            # Should go to the exception instead
            self.fail()
        except ValueError as e:
            self.assertTrue('bot' in str(e).lower())  # Indicator of "missing bot user something something"

    def test_persistent_token_exchanger_when_expired_token(self):
        factory = RequestFactory()
        request = factory.get('/', data={
            'type': 'setup',
            'data': self.VALID_DATA_JWT,
            'token': self.VALID_TOKEN_JWT
        })
        trigger = GiosgTriggerInDjango(request, SIGN_KEY)
        try:
            trigger.exchange_persistent_access_token()
            # Should go to the exception instead
            self.fail()
        except ValueError as e:
            self.assertTrue('expired' in str(e).lower())  # Indicator of "token expired something something"



from django.views import View
from django.http import HttpResponseBadRequest
from django.conf import settings
from ..lib.giosg_trigger_in_django import GiosgTriggerInDjango


class ApplicationTriggerView(View):
    """
    This module expects to find "GIOSG_APP_SECRET" variable in Django settings.
    This is intended to be subclassed, and used via "on_<trigger-type>" methods.
    """
    http_method_names = ['get']

    def get(self, request):
        # Leave validation to GiosgTriggerInDjango object
        try:
            trigger = GiosgTriggerInDjango(request, settings.GIOSG_APP_SECRET)
            handler = getattr(self, 'on_'+trigger.type, self.__unsupported_trigger_type)
            return handler(request, trigger)
        # Handle any giosg-auth-token validation errors
        except ValueError as e:
            return HttpResponseBadRequest(e)

    def __unsupported_trigger_type(self, *args, **kwargs):
        return HttpResponseBadRequest('Unsupported trigger type')

    def on_install(self, request, trigger):
        raise NotImplementedError

    def on_setup(self, request, trigger):
        raise NotImplementedError

    def on_uninstall(self, request, trigger):
        raise NotImplementedError

    def on_chat_start(self, request, trigger):
        raise NotImplementedError

    def on_chat_end(self, request, trigger):
        raise NotImplementedError

    def on_console_load(self, request, trigger):
        raise NotImplementedError

    def on_manual_dialog(self, request, trigger):
        raise NotImplementedError

    def on_manual_nav(self, request, trigger):
        raise NotImplementedError

    def on_chat_open(self, request, trigger):
        raise NotImplementedError

    def on_chat_close(self, request, trigger):
        raise NotImplementedError

    def on_chat_focus(self, request, trigger):
        raise NotImplementedError



import setuptools
from distutils.core import setup
with open('requirements.txt') as f:
  setup(
    name = 'giosgapps_bindings',         
    packages = setuptools.find_packages(),   
    version = '0.0.5',      
    license = 'MIT',        
    description = 'Module for Giosg Apps development',   
    author = 'Giosg',                  
    author_email = 'developers@giosg.com',      
    url = 'https://github.com/giosg/giosgapps_bindings',   
    download_url = 'https://github.com/giosg/giosgapps_bindings/archive/v0.0.4.tar.gz',    
    keywords = ['AUTH', 'GIOSG'],
    install_requires = list(f.read().splitlines()),  
    classifiers=[
      'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
      'Intended Audience :: Developers',      
      'Topic :: Software Development :: Build Tools',
      'License :: OSI Approved :: MIT License',  
      'Programming Language :: Python :: 3',   
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.6',
    ],
  )



from django.db import models


class GiosgAppInstallation(models.Model):
    """
    Store for persistent bot token, and any (protocol-specific) app-installation-specific data
    """
    installed_org_uuid = models.UUIDField()
    persistent_bot_token = models.TextField()
    persistent_token_prefix = models.CharField(max_length=255)

    class Meta:
        abstract = True
